import typing as T
import logging
from enum import Enum
from pydantic import BaseModel, PrivateAttr, model_validator, TypeAdapter
from pydantic_core import PydanticUndefined

from sqlalchemy.ext.asyncio import AsyncConnection
from sqlalchemy import RowMapping

from fastgql.query_builders.sql.query_builder import QueryBuilder, Cardinality

from asqlorm.sql.sql_alchemy import sql_pool, NoRowsReturnedError
from asqlorm.utils import TYPES_MAP

logger = logging.getLogger(__name__)


class raw_sql(BaseModel):
    stmt: str
    variables: dict[str, T.Any] | None = None


class FilterConnector(str, Enum):
    AND = " AND "
    OR = " OR "


class UpdateOperation(str, Enum):
    REPLACE = ":="
    ADD = "+="
    REMOVE = "-="


class UnsetError(Exception):
    pass


class ResolverError(Exception):
    pass


class NodeError(Exception):
    pass


VARS = dict[str, T.Any]
CONVERSION_FUNC = T.Callable[[str], T.Any]


IGNORE_SET: set[str] = {
    "__pydantic_validator__",
    "model_fields_set",
    "model_fields",
    "__pydantic_fields_set__",
}


class ConflictAction(str, Enum):
    DO_UPDATE = "DO UPDATE"
    DO_NOTHING = "DO NOTHING"


class Edge(BaseModel):
    from_: str
    cardinality: Cardinality
    is_required: bool


class EdgeWithResolver(Edge):
    resolver_type: T.Type["ResolverType"] | str


class ChildEdge(Edge):
    resolver: "ResolverType"


class ComputedColumn(BaseModel):
    path: str
    data_type: str = None
    is_nullable: bool

    basemodel_type: T.Type[BaseModel] | None = None
    cardinality: Cardinality = Cardinality.ONE

    custom_converter: T.Callable = None

    override: bool = None

    include_args_annotation: str = None
    update_qb: T.Callable[[QueryBuilder, T.Any], None] = None

    @model_validator(mode="after")
    def validate_column(self) -> "ComputedColumn":
        if not self.data_type and not self.custom_converter:
            raise ValueError(
                "You must provide either a data_type or a custom_converter."
            )
        if self.data_type and self.custom_converter:
            raise ValueError(
                "You cannot provide both a data_type or a custom_converter."
            )
        return self

    def __hash__(self):
        return hash(self.path)


class ComputedLink(BaseModel):
    node_name: str
    cardinality: Cardinality = Cardinality.ONE
    is_nullable: bool
    from_: str
    # from mapping for inheritance, like GiphyImage vs CloudinaryImage
    from_mapping: dict[str, str] | None = None

    annotations_by_kwarg: dict[str, str] = None
    update_qb: T.Callable[[QueryBuilder, T.Any], None] = None


class Node(BaseModel):
    def __getattribute__(self, name: str):
        if name not in IGNORE_SET and name not in self.model_fields_set:
            if name in self.model_fields:
                raise UnsetError(f"{name} is unset.")
        return super().__getattribute__(name)

    _children_cache: dict[str, "NodeType"] = PrivateAttr(default_factory=dict)

    _resolver_used: T.Optional["Resolver"] = PrivateAttr(None)

    __edges_map__: T.ClassVar[dict[str, EdgeWithResolver]]

    __computed_columns__: T.ClassVar[dict[str, ComputedColumn]]
    __computed_links__: T.ClassVar[dict[str, ComputedLink]]
    __table_name__: T.ClassVar[str]
    __primary_keys__: T.ClassVar[list[str]]

    _computed: dict[str, T.Any] = PrivateAttr(default_factory=dict)

    __columns_in_order__: T.ClassVar[list[str]]

    @property
    def computed(self) -> dict[str, T.Any]:
        return self._computed

    def _resolve_node(self, key: str) -> T.Union["NodeType", list["NodeType"]]:
        if key not in self._children_cache:
            raise NodeError(f"{key} is unset.")
        return self._children_cache[key]


NodeType = T.TypeVar("NodeType", bound=Node)


class Insert(BaseModel):
    pass


class Patch(BaseModel):
    pass


class Conflict(BaseModel):
    unique_index_str: str
    action: ConflictAction
    patch: T.Optional[Patch] = None
    columns_to_upsert: list[str] | None = None
    return_conflicting_rows: bool

    @model_validator(mode="after")
    def validate_action(self) -> "Conflict":
        if self.action == ConflictAction.DO_NOTHING and (
            self.patch or self.columns_to_upsert
        ):
            raise ValueError(
                "If the action is DO_NOTHING, do not give a patch or columns to upsert."
            )
        if self.action == ConflictAction.DO_UPDATE and not (
            self.patch or self.columns_to_upsert
        ):
            raise ValueError(
                "You must provide a patch or columns to upsert when the action is DO_UPDATE."
            )
        if (
            self.action == ConflictAction.DO_UPDATE
            and self.return_conflicting_rows is False
        ):
            raise ValueError(
                "If DO_UPDATE then it will always return conflicting rows, so return_conflicting_rows cannot be False."
            )
        return self

    def unique_column_names(self) -> list[str]:
        return [
            c_name.replace("(", "").replace(")", "").strip()
            for c_name in self.unique_index_str.split(",")
        ]


InsertType = T.TypeVar("InsertType", bound=Insert)


class Resolver(BaseModel):
    def __init__(self, /, **data: T.Any) -> None:  # type: ignore
        super().__init__(**data)
        self._qb = QueryBuilder(
            table_name=f'"{self._node.__table_name__}"',
            cardinality=Cardinality.ONE,  # This is dummy, change at build step
            from_=None,
        )
        for primary_key in self._node.__primary_keys__:
            self._qb.sel(name=primary_key, path=f"$current.{primary_key}")

    _node: T.ClassVar[T.Type[NodeType]]

    _qb: QueryBuilder = PrivateAttr

    _children: dict[str, "ChildEdge"] = PrivateAttr(default_factory=dict)

    _runtime__computed_columns__: dict[str, ComputedColumn] = PrivateAttr(
        default_factory=dict
    )

    def to_qb(
        self, cardinality: Cardinality, from_: str | None, replace_from: bool = True
    ) -> QueryBuilder:
        self._qb.cardinality = cardinality
        self._qb.set_from(from_=from_, replace_from=replace_from)
        return self._qb

    def build(
        self, cardinality: Cardinality, format_sql: bool = False, is_count: bool = False
    ) -> tuple[str, dict[T.Any]]:
        self._qb.cardinality = cardinality
        return self._qb.build_root(format_sql=format_sql, is_count=is_count)

    def add_query_variables(
        self: "ThisResolverType", variables: VARS | None
    ) -> "ThisResolverType":
        if variables:
            self._qb.add_variables(variables)
        return self

    def and_where(
        self: "ThisResolverType",
        filter_str: str,
        variables: VARS | None = None,
        replace_variables: bool = False,
    ) -> "ThisResolverType":
        self._qb.and_where(
            where=filter_str, variables=variables, replace_variables=replace_variables
        )
        return self

    def set_where(
        self: "ThisResolverType",
        where: str,
        variables: dict[str, T.Any] | None = None,
        replace_where: bool = False,
        replace_variables: bool = False,
    ) -> "ThisResolverType":
        self._qb.set_where(
            where=where,
            variables=variables,
            replace_where=replace_where,
            replace_variables=replace_variables,
        )
        return self

    def filter(
        self: "ThisResolverType", filter_str: str, variables: VARS | None = None
    ) -> "ThisResolverType":
        return self.and_where(filter_str=filter_str, variables=variables)

    def _filter_by(self: "ThisResolverType", **kwargs: T.Any) -> "ThisResolverType":
        kwargs = {k: v for k, v in kwargs.items() if v is not PydanticUndefined}
        if not kwargs:
            raise ResolverError("Nothing to filter by.")
        filter_strs = []
        variables = {}
        for field_name, field_value in kwargs.items():
            if computed_c := self._node.__computed_columns__.get(field_name):
                left_side = computed_c.path
            else:
                left_side = f"$current.{field_name}"
            if field_value is None:
                filter_strs.append(f"{left_side} IS NULL")
            else:
                filter_strs.append(f"{left_side} = ${field_name}")
                variables[field_name] = field_value
        filter_str = " AND ".join(sorted(filter_strs))
        return self.filter(filter_str=filter_str, variables=variables)

    def _filter_in(self: "ThisResolverType", **kwargs: T.Any) -> "ThisResolverType":
        kwargs = {k: v for k, v in kwargs.items() if v is not PydanticUndefined}
        if not kwargs:
            raise ResolverError("Nothing to filter by.")
        filter_strs = []
        variables = {}
        for field_name, value_lst in kwargs.items():
            if computed_c := self._node.__computed_columns__.get(field_name):
                left_side = computed_c.path
            else:
                left_side = f"$current.{field_name}"
            filter_strs.append(f"{left_side} = ANY(${field_name})")
            variables[field_name] = value_lst
        filter_str = " AND ".join(sorted(filter_strs))
        return self.filter(filter_str=filter_str, variables=variables)

    def order_by(
        self: "ThisResolverType", order_by_str: str, variables: VARS | None = None
    ) -> "ThisResolverType":
        self._qb.set_order_by(order_by=order_by_str, variables=variables)
        return self

    def offset(self: "ThisResolverType", /, _: int | None) -> "ThisResolverType":
        self._qb.set_offset(offset=_)
        return self

    def limit(self: "ThisResolverType", /, _: int | None) -> "ThisResolverType":
        self._qb.set_limit(limit=_)
        return self

    def include_fields(
        self: "ThisResolverType", *fields_to_include: str
    ) -> "ThisResolverType":
        for f in fields_to_include:
            if computed_col := self._node.__computed_columns__.get(f):
                self._qb.sel(name=f, path=computed_col.path)
            else:
                self._qb.sel(f)
        return self

    def _include(self: "ThisResolverType", **kwargs) -> "ThisResolverType":
        fields_to_include: list[str] = []
        for k, v in kwargs.items():
            if v is False or v is PydanticUndefined:
                continue
            if v is not True:
                # must be computed update qb
                self._node.__computed_columns__[k].update_qb(self._qb, *v)
            fields_to_include.append(k)
        return self.include_fields(*fields_to_include)

    def include_all_columns(
        self: "ThisResolverType",
        include_computed_columns: bool = False,
        exclude: set[str] | None = None,
        variables: dict[str, T.Any] | None = None,
    ) -> "ThisResolverType":
        cols = self._node.model_fields.keys()
        if not include_computed_columns:
            cols = cols - {
                k for k, v in self._node.__computed_columns__.items() if not v.override
            }
        if exclude:
            cols = cols - exclude
        if variables:
            self.add_query_variables(variables)
        self.include_fields(*cols)
        return self

    def add_child(
        self: "ThisResolverType",
        key: str,
        resolver: "ResolverType",
        cardinality: Cardinality,
        from_: str,
        is_required: bool,
    ) -> "ThisResolverType":
        if key in self._children:
            raise ResolverError(f"Key {key} is already set.")
        child_edge = ChildEdge(
            resolver=resolver,
            cardinality=cardinality,
            from_=from_,
            is_required=is_required,
        )
        self._children[key] = child_edge
        child_qb = child_edge.resolver.to_qb(
            cardinality=child_edge.cardinality, from_=child_edge.from_
        )
        self._qb.sel_sub(key, child_qb)
        return self

    def add_computed(
        self: "ThisResolverType", name: str, computed_column: ComputedColumn
    ) -> "ThisResolverType":
        if name in self._runtime__computed_columns__:
            raise ResolverError(f"Computed column {name} is already given.")
        self._runtime__computed_columns__[name] = computed_column
        self._qb.sel(name=name, path=computed_column.path)
        return self

    def computed(
        self: "ThisResolverType",
        *,
        name: str,
        path: str,
        data_type: str = None,
        is_nullable: bool,
        custom_converter: T.Callable = None,
    ) -> "ThisResolverType":
        if name in self._runtime__computed_columns__:
            raise ResolverError(f"Computed column {name} is already given.")
        computed_kwargs = {
            k: v
            for k, v in {
                "path": path,
                "data_type": data_type,
                "is_nullable": is_nullable,
                "custom_converter": custom_converter,
            }.items()
            if v is not None
        }
        computed_column = ComputedColumn(**computed_kwargs)
        return self.add_computed(name=name, computed_column=computed_column)

    async def _get_node(
        self,
        field_name: str,
        value: T.Any,
        format_sql: bool = False,
        conn: AsyncConnection | None = None,
    ) -> T.Optional[NodeType]:
        if isinstance(value, tuple):
            keys = field_name.split("__")
            if len(keys) != len(value):
                raise ResolverError(
                    f"Keys length: {keys=} does not match tuple length: {value=}"
                )
            left_side_strs: list[str] = []
            right_side_strs: list[str] = []
            variables: dict[str, T.Any] = {}
            for key, val in zip(keys, value):
                left_side_strs.append(f"$current.{key}")
                right_side_strs.append(f"${key}")
                variables[key] = val

            self.filter(
                f"({', '.join(left_side_strs)}) = ({', '.join(right_side_strs)})",
                variables,
            )
        else:
            self._filter_by(**{field_name: value}).limit(1)
        s, v = self.build(cardinality=Cardinality.ONE, format_sql=format_sql)
        raw_val = await sql_pool.query_val(s, v, log_query=False, conn=conn)
        if raw_val is None:
            return None
        return self.parse_node(**raw_val)

    async def _get(
        self, conn: AsyncConnection | None = None, **kwargs
    ) -> T.Optional[NodeType]:
        field_name, value = validate_get_kwargs(**kwargs)
        return await self._get_node(field_name=field_name, value=value, conn=conn)

    async def _gerror(self, conn: AsyncConnection | None = None, **kwargs) -> NodeType:
        field_name, value = validate_get_kwargs(**kwargs)
        node = await self._get_node(field_name=field_name, value=value, conn=conn)
        if not node:
            raise ResolverError(
                f"{self._node.__name__} not found where {field_name} == {value}."
            )
        return node

    async def count(
        self,
        *,
        format_sql: bool = False,
        log_query: bool = False,
        conn: AsyncConnection | None = None,
    ) -> int:
        s, v = self.build(
            cardinality=Cardinality.MANY, format_sql=format_sql, is_count=True
        )
        return await sql_pool.query_val(s, v, log_query=log_query, conn=conn)

    async def _query(
        self,
        *,
        format_sql: bool = False,
        log_query: bool = False,
        conn: AsyncConnection | None = None,
    ) -> list[NodeType]:
        s, v = self.build(cardinality=Cardinality.MANY, format_sql=format_sql)
        raw_list = await sql_pool.query_val(s, v, log_query=log_query, conn=conn)
        return [self.parse_node(**item) for item in raw_list]

    async def first_or_none(
        self,
        *,
        format_sql: bool = False,
        log_query: bool = False,
        conn: AsyncConnection | None = None,
    ) -> T.Optional[NodeType]:
        s, v = self.limit(1).build(cardinality=Cardinality.MANY, format_sql=format_sql)
        raw_list = await sql_pool.query_val(s, v, log_query=log_query, conn=conn)
        if not raw_list:
            return None
        return self.parse_node(**raw_list[0])

    async def first(
        self,
        *,
        format_sql: bool = False,
        log_query: bool = False,
        conn: AsyncConnection | None = None,
    ) -> NodeType:
        node = await self.first_or_none(
            format_sql=format_sql, log_query=log_query, conn=conn
        )
        if not node:
            raise ResolverError("No value found.")
        return node

    async def one_or_none(
        self,
        *,
        format_sql: bool = False,
        log_query: bool = False,
        conn: AsyncConnection | None = None,
    ) -> T.Optional[NodeType]:
        s, v = self.limit(2).build(cardinality=Cardinality.MANY, format_sql=format_sql)
        raw_list = await sql_pool.query_val(s, v, log_query=log_query, conn=conn)
        raw_list_len = len(raw_list)
        if raw_list_len == 0:
            return None
        if raw_list_len > 1:
            raise ResolverError("More than one result returned.")
        return self.parse_node(**raw_list[0])

    async def one(
        self,
        *,
        format_sql: bool = False,
        log_query: bool = False,
        conn: AsyncConnection | None = None,
    ) -> NodeType:
        node = await self.one_or_none(
            format_sql=format_sql, log_query=log_query, conn=conn
        )
        if not node:
            raise ResolverError("No value found.")
        return node

    def parse_mutation_model(
        self, model: Insert | Patch, prefix: str | None = None
    ) -> tuple[dict[str, str], dict[str, T.Any]]:
        # mode json is incorrect for date-times
        dump_d = model.model_dump(mode="python", exclude_unset=True)
        field_names_to_value_strs: dict[str, str] = {}
        variables: dict[str, T.Any] = {}
        for k, dump_v in dump_d.items():
            model_v = getattr(model, k)
            if isinstance(model_v, raw_sql):
                if model_v.variables:
                    for raw_k, raw_v in model_v.variables.items():
                        if raw_k in variables and raw_v != variables[raw_k]:
                            raise ResolverError(
                                f"Variables {raw_k} is already being used."
                            )
                        variables[raw_k] = raw_v
                field_names_to_value_value = model_v.stmt
            else:
                v = dump_v
                if computed_c := self._node.__computed_columns__.get(k):
                    if computed_c.basemodel_type:
                        if model_v is not None:
                            if computed_c.cardinality == Cardinality.ONE:
                                v = model_v.model_dump_json()
                            else:
                                v = (
                                    TypeAdapter(list[computed_c.basemodel_type])
                                    .dump_json(model_v)
                                    .decode("utf-8")
                                )
                if prefix:
                    variable_name = f"{prefix}{k}"
                else:
                    variable_name = k
                if variable_name in variables and v != variables[variable_name]:
                    raise ResolverError(
                        f"Variable '{variable_name}' is already being used."
                    )
                variables[variable_name] = v
                field_names_to_value_value = f":{variable_name}"
            field_names_to_value_strs[k] = field_names_to_value_value
        return field_names_to_value_strs, variables

    def build_conflict_str(
        self, conflict: Conflict | None, existing_variables: dict[str, T.Any] | None
    ) -> tuple[str, dict[str, T.Any]]:
        if existing_variables is None:
            existing_variables = {}
        conflict_str = ""
        if conflict:
            conflict_str = (
                f"ON CONFLICT {conflict.unique_index_str} {conflict.action.value}"
            )
            conflict_strs: list[str] = []
            if conflict.patch:
                _field_names_to_value_strs, _variables = self.parse_mutation_model(
                    model=conflict.patch, prefix="_patch_"
                )
                for _k, _v in _variables.items():
                    if _k in existing_variables and _v != existing_variables[_k]:
                        raise ResolverError(
                            f"Conflict variable '{_k}' is already being used."
                        )
                    existing_variables[_k] = _v
                for _field_name, _value in _field_names_to_value_strs.items():
                    conflict_strs.append(f"{_field_name} = {_value}")
            if conflict.columns_to_upsert:
                for col in conflict.columns_to_upsert:
                    conflict_strs.append(f"{col} = excluded.{col}")
            if conflict_strs:
                conflict_str = f'{conflict_str} SET {", ".join(conflict_strs)}'
        return conflict_str, existing_variables

    def build_insert_q_and_v(
        self,
        insert: Insert,
        conflict: Conflict | None = None,
    ) -> tuple[str, dict[str, T.Any]]:
        field_names_to_value_strs, variables = self.parse_mutation_model(model=insert)

        conflict_str, variables = self.build_conflict_str(
            conflict=conflict, existing_variables=variables
        )

        if (
            conflict_str
            and conflict.action == ConflictAction.DO_NOTHING
            and conflict.return_conflicting_rows
        ):
            conflict_col_names = conflict.unique_column_names()
            unique_col_names_strs: list[str] = []
            for n in conflict_col_names:
                if n not in field_names_to_value_strs:
                    raise ResolverError(
                        f"Cannot return conflicting rows on {n} since {n} was not given for the insert."
                    )
                unique_col_names_strs.append(field_names_to_value_strs[n])
            unique_col_names_str = ", ".join(
                [field_names_to_value_strs[c_name] for c_name in conflict_col_names]
            )
            if len(conflict_col_names) > 1:
                unique_col_names_str = f"({unique_col_names_str})"
            where_unique_index_str = conflict.unique_index_str
            if len(conflict_col_names) == 1:
                where_unique_index_str = conflict.unique_index_str.replace(
                    "(", ""
                ).replace(")", "")
            conflicting_from_str = f"{self._table_name} WHERE {where_unique_index_str} = {unique_col_names_str}"
            complex_return_str = self._returning_str_complex(use_table_name=False)
            return_select_str = f"""
, combined AS (
    SELECT {complex_return_str} FROM inserted
    UNION ALL
    SELECT {complex_return_str} FROM {conflicting_from_str}
    AND NOT EXISTS (SELECT 1 FROM inserted)
)
SELECT {complex_return_str} FROM combined;
            """
        else:
            return_select_str = "SELECT * FROM inserted"

        field_names = ", ".join(field_names_to_value_strs.keys())
        values = ", ".join(field_names_to_value_strs.values())

        q = f"""
WITH inserted AS (
    INSERT INTO {self._table_name} ({field_names})
    VALUES ({values})
    {conflict_str}
    RETURNING {self._returning_str}
)
{return_select_str}
            """

        return q, variables

    def build_insert_many_q_and_v(
        self, inserts: list[Insert], conflict: Conflict | None = None
    ) -> tuple[str, T.Any]:
        if self._qb.where or self._qb.from_:
            raise ResolverError("Resolver cannot have filters set for inserts.")

        prev_field_names_to_value_strs: dict[str, str] | None = None
        field_names_to_value_strs_list: list[dict[str, str]] = []
        variables_list: list[dict[str, T.Any]] = []
        for insert in inserts:
            # do not need to check that variables have the same keys since that is taken care of
            # with prev_field_names_to_value_strs comparisons
            field_names_to_value_strs, variables = self.parse_mutation_model(
                model=insert
            )
            if prev_field_names_to_value_strs is None:
                prev_field_names_to_value_strs = field_names_to_value_strs
            else:
                if field_names_to_value_strs != prev_field_names_to_value_strs:
                    raise ResolverError(
                        "All inserts must have the same fields to insert, including the same raw_sql."
                    )

            field_names_to_value_strs_list.append(field_names_to_value_strs)
            variables_list.append(variables)

        # first, build insert string
        fields_name_strs: list[str] = []
        select_val_strs: list[str] = []
        first_field_names_to_value_strs = field_names_to_value_strs_list[0]
        first_variables = variables_list[0]
        for col in self._node.__columns_in_order__:
            _values: list[T.Any] = []
            if col in first_field_names_to_value_strs:
                fields_name_strs.append(col)
                if col in first_variables:
                    select_val_strs.append(f"o.{col}")
                else:
                    # if it isn't a variable, add the value (from raw_sql) directly
                    select_val_strs.append(first_field_names_to_value_strs[col])

        patch_variables = {}
        conflict_str, patch_variables = self.build_conflict_str(
            conflict=conflict, existing_variables=patch_variables
        )

        return_select_str = "SELECT * FROM inserted"

        if (
            conflict_str
            and conflict.action == ConflictAction.DO_NOTHING
            and conflict.return_conflicting_rows
        ):
            # this means that you will have to return a UNION of the conflicts AND the returnings
            conflict_column_names = conflict.unique_column_names()
            conflict_col_names_str = ", ".join(
                [f"o.{n}" for n in conflict_column_names]
            )
            where_unique_index_str = conflict.unique_index_str
            if len(conflict_column_names) == 1:
                where_unique_index_str = conflict.unique_index_str.replace(
                    "(", ""
                ).replace(")", "")
            conflicting_from_str = f"{self._table_name} WHERE {where_unique_index_str} in (SELECT {conflict_col_names_str} FROM unnest($1::{self._table_name}[]) as o)) UNION ({return_select_str})"
            complex_return_str = self._returning_str_complex(use_table_name=False)
            return_select_str = f"""
, combined AS (
    SELECT {complex_return_str} FROM inserted
    UNION ALL
    SELECT {complex_return_str} FROM {conflicting_from_str}
    AND NOT EXISTS (SELECT 1 FROM inserted)
)
SELECT {complex_return_str} FROM combined;
                        """

        q = f"""
WITH inserted as (
    INSERT INTO {self._table_name} ({', '.join(fields_name_strs)})
    (SELECT {', '.join(select_val_strs)} FROM unnest((:_raw_models)::{self._table_name}[]) as o)
    {conflict_str}
    RETURNING {self._returning_str}
)
{return_select_str}
            """

        values: list[tuple[T.Any, ...]] = []
        for field_names_to_value_strs, variables in zip(
            field_names_to_value_strs_list, variables_list
        ):
            _values: list[T.Any] = []
            for col in self._node.__columns_in_order__:
                if col in field_names_to_value_strs and col in variables:
                    _v = variables[col]
                else:
                    _v = None
                _values.append(_v)

            values.append(tuple(_values))

        return q, {"_raw_models": values, **patch_variables}

    def build_patch_q_and_v(
        self, patch: Patch, patch_all: bool
    ) -> tuple[str, dict[str, T.Any]]:
        field_names_to_value_strs, variables = self.parse_mutation_model(model=patch)
        set_strs: list[str] = []

        for field_name, value in field_names_to_value_strs.items():
            set_strs.append(f"{field_name} = {value}")

        if not patch_all:
            if not self._qb.where:
                raise ResolverError(
                    "Unless you want to patch all rows (patch_all=True), you need to give a WHERE clause."
                )

        where_s = self._qb.replace_current_and_parent(
            s=f"WHERE {self._qb.where}",
            table_alias=self._table_name,
            parent_table_alias=None,
        )
        where_s, where_v = self._qb.prepare_query_sqlalchemy(
            sql=where_s, params=self._qb.variables
        )
        for k, v in where_v.items():
            if k in variables and v != variables[k]:
                raise ResolverError(f"Variables {k} is already being used.")
            variables[k] = v

        q = f"""
UPDATE {self._table_name} SET {', '.join(set_strs)}
{where_s.replace('$current.', f'{self._table_name}.')}
RETURNING {self._returning_str}
            """

        return q, variables

    def build_delete_q_and_v(self, delete_all: bool) -> tuple[str, dict[str, T.Any]]:
        if not delete_all:
            if not self._qb.where:
                raise ResolverError(
                    "Unless you want to delete all rows (delete_all=True), you need to give a WHERE clause."
                )

        where_s = self._qb.replace_current_and_parent(
            s=f"WHERE {self._qb.where}",
            table_alias=self._table_name,
            parent_table_alias=None,
        )
        where_s, where_v = self._qb.prepare_query_sqlalchemy(
            sql=where_s, params=self._qb.variables
        )

        q = f"""
DELETE FROM {self._table_name}
{where_s.replace('$current.', f'{self._table_name}.')}
RETURNING {self._returning_str}
                """

        return q, where_v

    @property
    def _table_name(self) -> str:
        return f'"{self._node.__table_name__}"'

    @property
    def _returning_str(self) -> str:
        """assume primary keys are not computed or complex paths"""
        returning_str = "*"
        returning_strs: set[str] = set()
        # include fields to include as long as the path == name or $current.name
        for primary_key in self._node.__primary_keys__:
            returning_strs.add(f"{self._table_name}.{primary_key} as {primary_key}")
        for sel in self._qb.selections:
            if sel.is_simple_column:
                returning_strs.add(f"{self._table_name}.{sel.name} as {sel.name}")
        if returning_strs:
            returning_str = ", ".join(sorted(returning_strs))
        return returning_str

    def _returning_str_complex(self, use_table_name: bool) -> str:
        returning_str = "*"
        returning_strs: set[str] = set()
        # include fields to include as long as the path == name or $current.name
        for primary_key in self._node.__primary_keys__:
            if use_table_name:
                returning_strs.add(f"{self._table_name}.{primary_key} as {primary_key}")
            else:
                returning_strs.add(primary_key)
        for sel in self._qb.selections:
            if sel.is_simple_column:
                if use_table_name:
                    returning_strs.add(f"{self._table_name}.{sel.name} as {sel.name}")
                else:
                    returning_strs.add(sel.name)
        if returning_strs:
            returning_str = ", ".join(sorted(returning_strs))
        return returning_str

    async def node_from_raw_val(
        self,
        raw_val: RowMapping,
        conn: AsyncConnection | None,
        format_sql: bool,
        commit_after: bool,
        force_refresh: bool,
        error_if_need_to_refresh: bool = False,
    ) -> NodeType:
        if conn and commit_after:
            logger.debug("[COMMIT AFTER]")
            await conn.commit()
        # for now, if there are selections other than id, refresh
        should_refresh = force_refresh
        if not should_refresh:
            for s in self._qb.selections:
                if s.name not in raw_val or not s.is_simple_column:
                    should_refresh = True
                    break
        if should_refresh:
            if error_if_need_to_refresh:
                raise ResolverError(
                    "This resolver requires a refresh but the node was deleted."
                )
            logger.debug("[REFRESH AFTER]")
            # what about composite primary keys? Or no primary keys?
            kwargs = {k: raw_val[k] for k in self._node.__primary_keys__}
            return await self._gerror(conn=conn, **kwargs)
        else:
            parsed_d: dict[str, T.Any] = {}
            for k, v in raw_val.items():
                if k not in self._node.__computed_columns__:
                    parsed_d[k] = v
            return self.parse_node(**parsed_d)

    async def nodes_from_raw_vals(
        self,
        raw_vals: T.Sequence[RowMapping],
        conn: AsyncConnection | None,
        format_sql: bool,
        commit_after: bool,
        force_refresh: bool,
        error_if_need_to_refresh: bool = False,
    ) -> list[NodeType]:
        if conn and commit_after:
            logger.debug("[COMMIT AFTER]")
            await conn.commit()
        # for now, if there are selections other than id, refresh
        should_refresh = force_refresh
        if not should_refresh:
            for raw_val in raw_vals:
                for s in self._qb.selections:
                    if s.name not in raw_val or not s.is_simple_column:
                        should_refresh = True
                        break
        if should_refresh:
            if error_if_need_to_refresh:
                raise ResolverError(
                    "This resolver requires a refresh but the node was deleted."
                )
            logger.debug("[REFRESH AFTER]")
            # what about composite primary keys? Or no primary keys?
            kwargs = {
                k: [raw_val[k] for raw_val in raw_vals]
                for k in self._node.__primary_keys__
            }
            return await self._filter_in(**kwargs)._query(
                format_sql=format_sql, conn=conn, log_query=True
            )
        else:
            nodes: list[NodeType] = []
            for raw_val in raw_vals:
                parsed_d: dict[str, T.Any] = {}
                for k, v in raw_val.items():
                    if k not in self._node.__computed_columns__:
                        parsed_d[k] = v
                nodes.append(self.parse_node(**parsed_d))
            return nodes

    async def insert_one_or_none(
        self,
        insert: Insert,
        *,
        format_sql: bool = False,
        log_query: bool = False,
        conn: AsyncConnection | None = None,
        commit_after_insert: bool = False,
        force_refresh: bool = False,
        conflict: Conflict | None = None,
    ) -> NodeType | None:
        q, v = self.build_insert_q_and_v(insert=insert, conflict=conflict)
        raw_val = await sql_pool.query_record(q, v, log_query=log_query, conn=conn)
        if conn and commit_after_insert:
            logger.debug("insert committing")
            await conn.commit()
        if not raw_val:
            return None
        return await self.node_from_raw_val(
            raw_val=raw_val,
            conn=conn,
            format_sql=format_sql,
            commit_after=commit_after_insert,
            force_refresh=force_refresh,
        )

    async def insert_one(
        self,
        insert: Insert,
        *,
        format_sql: bool = False,
        log_query: bool = False,
        conn: AsyncConnection | None = None,
        commit_after_insert: bool = False,
        force_refresh: bool = False,
        conflict: Conflict | None = None,
    ) -> NodeType:
        if conflict and conflict.return_conflicting_rows is not True:
            raise ResolverError(
                "If you want to insert_one with a conflict, conflict.return_conflicting_rows must be True. "
                "Otherwise, choose insert_one_or_none."
            )
        node = await self.insert_one_or_none(
            insert=insert,
            format_sql=format_sql,
            log_query=log_query,
            conn=conn,
            commit_after_insert=commit_after_insert,
            force_refresh=force_refresh,
            conflict=conflict,
        )
        if not node:
            raise ResolverError("Nothing was inserted.")
        return node

    async def insert_many(
        self,
        inserts: list[Insert],
        conflict: Conflict | None = None,
        *,
        format_sql: bool = False,
        log_query: bool = False,
        conn: AsyncConnection | None = None,
        commit_after_insert: bool = False,
        force_refresh: bool = False,
    ) -> list[NodeType]:
        """If you want to try returning in the same query, see the deleted code that tried to do it:
        ff4491b..ad721c4  main -> main"""
        if not inserts:
            return []
        q, v = self.build_insert_many_q_and_v(inserts=inserts, conflict=conflict)
        raw_vals = await sql_pool.query_records(q, v, log_query=log_query, conn=conn)
        return await self.nodes_from_raw_vals(
            raw_vals=raw_vals,
            conn=conn,
            format_sql=format_sql,
            commit_after=commit_after_insert,
            force_refresh=force_refresh,
        )

    async def patch_many(
        self,
        patch: Patch,
        *,
        format_sql: bool = False,
        log_query: bool = False,
        conn: AsyncConnection | None = None,
        commit_after_patch: bool = False,
        force_refresh: bool = False,
        patch_all: bool = False,
    ) -> list[NodeType]:
        q, v = self.build_patch_q_and_v(patch=patch, patch_all=patch_all)
        raw_vals = await sql_pool.query_records(q, v, log_query=log_query, conn=conn)
        return await self.nodes_from_raw_vals(
            raw_vals=raw_vals,
            conn=conn,
            format_sql=format_sql,
            commit_after=commit_after_patch,
            force_refresh=force_refresh,
        )

    async def patch_one(
        self,
        patch: Patch,
        *,
        format_sql: bool = False,
        log_query: bool = False,
        conn: AsyncConnection | None = None,
        commit_after_patch: bool = False,
        force_refresh: bool = False,
    ) -> NodeType:
        q, v = self.build_patch_q_and_v(patch=patch, patch_all=False)
        if conn:
            raw_val = await sql_pool.query_records(
                q, v, log_query=log_query, conn=conn, confirm_only_one=True
            )
        else:
            raw_val = await sql_pool.query_records(
                q, v, log_query=log_query, confirm_only_one=True
            )
        return await self.node_from_raw_val(
            raw_val=raw_val,
            conn=conn,
            format_sql=format_sql,
            commit_after=commit_after_patch,
            force_refresh=force_refresh,
        )

    async def delete_one_or_none(
        self,
        *,
        format_sql: bool = False,
        log_query: bool = False,
        conn: AsyncConnection | None = None,
        commit_after_delete: bool = False,
    ) -> NodeType | None:
        q, v = self.build_delete_q_and_v(delete_all=False)
        try:
            if conn:
                raw_val = await sql_pool.query_records(
                    q, v, log_query=log_query, conn=conn, confirm_only_one=True
                )
            else:
                raw_val = await sql_pool.query_records(
                    q, v, log_query=log_query, confirm_only_one=True
                )
        except NoRowsReturnedError:
            return None
        return await self.node_from_raw_val(
            raw_val=raw_val,
            conn=conn,
            format_sql=format_sql,
            commit_after=commit_after_delete,
            force_refresh=False,
            error_if_need_to_refresh=True,
        )

    async def delete_one(
        self,
        *,
        format_sql: bool = False,
        log_query: bool = False,
        conn: AsyncConnection | None = None,
        commit_after_delete: bool = False,
    ) -> NodeType:
        row = await self.delete_one_or_none(
            format_sql=format_sql,
            log_query=log_query,
            conn=conn,
            commit_after_delete=commit_after_delete,
        )
        if not row:
            raise ResolverError("No rows were deleted.")
        return row

    async def delete_many(
        self,
        *,
        format_sql: bool = False,
        log_query: bool = False,
        conn: AsyncConnection | None = None,
        commit_after_delete: bool = False,
        delete_all: bool = False,
    ) -> list[NodeType]:
        q, v = self.build_delete_q_and_v(delete_all=delete_all)
        raw_vals = await sql_pool.query_records(q, v, log_query=log_query, conn=conn)
        return await self.nodes_from_raw_vals(
            raw_vals=raw_vals,
            conn=conn,
            format_sql=format_sql,
            commit_after=commit_after_delete,
            force_refresh=False,
            error_if_need_to_refresh=True,
        )

    def parse_node(self, **kwargs) -> NodeType:
        node = self._node(**kwargs)
        # do computeds here
        if self._runtime__computed_columns__:
            for name, col in self._runtime__computed_columns__.items():
                if name not in kwargs:
                    raise ResolverError(f"Computed column {name} not found in results.")
                computed_val = kwargs[name]
                if computed_val is None:
                    if not col.is_nullable:
                        raise ResolverError(
                            f"Computed column {name} is not nullable but was null."
                        )
                else:
                    if col.custom_converter:
                        computed_val = col.custom_converter(computed_val)
                    else:
                        if col.data_type not in TYPES_MAP:
                            raise ResolverError(
                                f"You must either give a custom converter or a valid data_type for computed column {name}."
                            )
                        type_map_val = TYPES_MAP[col.data_type]
                        if col.cardinality == Cardinality.ONE:
                            if not isinstance(computed_val, type_map_val.val):
                                computed_val = type_map_val.serializer(computed_val)
                        else:
                            computed_val_list = []
                            for v in computed_val:
                                if not isinstance(v, type_map_val.val):
                                    v = type_map_val.serializer(v)
                                computed_val_list.append(v)
                            computed_val = computed_val_list
                node.computed[name] = computed_val

        # now do nested
        for key, edge in self._children.items():
            raw_child = kwargs[key]
            if edge.cardinality == Cardinality.ONE:
                if raw_child is None:
                    if edge.is_required:
                        raise ResolverError(f"{key} is required but is null.")
                    node._children_cache[key] = None
                else:
                    node._children_cache[key] = edge.resolver.parse_node(**raw_child)
            else:
                node._children_cache[key] = [
                    edge.resolver.parse_node(**d) for d in raw_child
                ]

        # TODO extras later

        return node


def validate_get_kwargs(**kwargs) -> tuple[str, T.Any]:
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    kwargs_len = len(kwargs)
    if kwargs_len != 1:
        raise ResolverError(
            f"Must only give one argument, received {kwargs_len} arguments: {kwargs}."
        )
    field_name, value = list(kwargs.items())[0]
    return field_name, value


ResolverType = T.TypeVar("ResolverType", bound=Resolver)
ThisResolverType = ResolverType



import typing as T
import re
from black import format_str, FileMode
from pydantic import BaseModel, Field
import psycopg

from asqlorm.utils import TYPES_MAP
from asqlorm.models import (
    Node,
    Resolver,
    ComputedColumn,
    Cardinality,
    ComputedLink,
)


class GeneratorError(Exception):
    pass


class TableConfig(BaseModel):
    base_node: T.Type[Node] | None = None
    base_resolver: T.Type[Resolver] | None = None
    # columns_to_default_include: set[str] = Field(default_factory=set)


class Ref(BaseModel):
    table_name: str
    column_name: str
    is_nullable: bool

    @classmethod
    def from_str(cls, s: str, is_nullable: bool):
        ref_table_name = s[: s.index("(")]
        ref_col_name = s[s.index("(") + 1 : s.index(")")]
        return cls(
            table_name=ref_table_name, column_name=ref_col_name, is_nullable=is_nullable
        )


DEFAULT_ENUMS_QUERY: str = """
SELECT n.nspname   AS schema_name,
       t.typname   AS enum_type,
       e.enumlabel AS enum_value
FROM pg_type t
         JOIN
     pg_enum e ON t.oid = e.enumtypid
         JOIN
     pg_catalog.pg_namespace n ON n.oid = t.typnamespace
WHERE n.nspname = 'public'
ORDER BY t.typname,
         e.enumsortorder;
"""

DEFAULT_COLUMNS_QUERY: str = """
SELECT col.table_schema,
       col.table_name,
       col.column_name,
       CASE
           WHEN col.data_type = 'ARRAY' THEN array_type.typname || '[]'
           WHEN col.data_type = 'USER-DEFINED' THEN pg_type.typname
           ELSE col.data_type
       END as data_type,
       CASE
           WHEN col.is_nullable = 'YES' THEN true
           ELSE false
       END as is_nullable,
       col.column_default as default_value, -- Added default value column
       (SELECT string_agg(DISTINCT cons_info, '; ')
        FROM (
              SELECT tc.constraint_type || ': ' || tc.constraint_name ||
                     CASE
                         WHEN tc.constraint_type = 'FOREIGN KEY'
                             THEN ' references ' || ccu.table_name || '(' || ccu.column_name || ')'
                         ELSE ''
                         END AS cons_info
              FROM information_schema.table_constraints tc
                       JOIN information_schema.key_column_usage kcu ON tc.constraint_name = kcu.constraint_name
                       AND tc.table_schema = kcu.table_schema
                       AND tc.table_name = kcu.table_name
                       LEFT JOIN information_schema.constraint_column_usage ccu
                             ON ccu.constraint_name = tc.constraint_name
                             AND ccu.table_schema = tc.table_schema
              WHERE kcu.column_name = col.column_name
                AND kcu.table_name = col.table_name
                AND kcu.table_schema = col.table_schema

              UNION ALL

              SELECT 'UNIQUE INDEX: ' || idx.relname AS cons_info
              FROM pg_index
                       JOIN pg_class idx ON pg_index.indexrelid = idx.oid
                       JOIN pg_class tbl ON pg_index.indrelid = tbl.oid
                       JOIN pg_namespace ns ON tbl.relnamespace = ns.oid
                       JOIN pg_attribute attr ON attr.attrelid = tbl.oid AND attr.attnum = ANY(pg_index.indkey)
              WHERE pg_index.indisunique = true
                AND ns.nspname = col.table_schema
                AND tbl.relname = col.table_name
                AND attr.attname = col.column_name
                AND (SELECT COUNT(*) FROM pg_attribute WHERE attrelid = tbl.oid AND attnum = ANY(pg_index.indkey)) = 1
        ) AS detailed_constraints) AS constraints_info
FROM information_schema.columns col
         LEFT JOIN pg_type ON col.udt_name = pg_type.typname
         LEFT JOIN pg_type array_type ON pg_type.typelem = array_type.oid
WHERE col.table_schema NOT IN ('pg_catalog', 'information_schema')
ORDER BY col.table_schema, col.table_name, col.ordinal_position;
"""

DEFAULT_TABLES_QUERY = """
SELECT
    n.nspname AS schema_name,
    c.relname AS table_name,
    d.description AS comment,
    (
        SELECT string_agg(cons.constr_details, '; ')
        FROM (
            SELECT
                CASE pc.contype
                    WHEN 'p' THEN 'PRIMARY KEY (' || pg_get_constraintdef(pc.oid) || ')'
                    WHEN 'f' THEN 'FOREIGN KEY (' || pg_get_constraintdef(pc.oid) || ')'
                    WHEN 'u' THEN 'UNIQUE (' || pg_get_constraintdef(pc.oid) || ')'
                    WHEN 'c' THEN 'CHECK (' || pg_get_constraintdef(pc.oid) || ')'
                    WHEN 'x' THEN 'EXCLUDE (' || pg_get_constraintdef(pc.oid) || ')'
                    ELSE 'OTHER'
                END AS constr_details
            FROM pg_constraint pc
            WHERE pc.conrelid = c.oid
        ) cons
    ) AS constraints
FROM
    pg_class c
JOIN
    pg_namespace n ON c.relnamespace = n.oid
LEFT JOIN
    pg_description d ON c.oid = d.objoid
WHERE
    c.relkind = 'r' -- 'r' indicates a regular table
    AND n.nspname NOT IN ('pg_catalog', 'information_schema')
ORDER BY
    n.nspname,
    c.relname;
"""


def convert_to_variable_name(enum_name: str) -> str:
    return re.sub(r"\W|^(?=\d)", "_", enum_name)


def build_enums_str(
    *,
    db_dsn: str,
    custom_query: str = DEFAULT_ENUMS_QUERY,
    custom_enum_converter: T.Callable[[str], str] = None,
) -> str:
    enum_converter = custom_enum_converter or convert_to_variable_name
    enums_d: dict[str, list[str]] = {}
    with psycopg.connect(db_dsn) as conn:
        conn.row_factory = psycopg.rows.dict_row
        with conn.cursor() as cur:
            cur.execute(custom_query)
            for record in cur.fetchall():
                enum_type = record["enum_type"]
                if enum_type not in enums_d:
                    enums_d[enum_type] = []
                enums_d[enum_type].append(record["enum_value"])
    # now build the python string from this
    enum_strs: list[str] = []
    for enum_name, enum_values in enums_d.items():
        enum_values_str = "\n".join(
            [f'\t{enum_converter(v)} = "{v}"' for v in enum_values]
        )
        enum_strs.append(
            f"""
class {enum_name}(str, Enum):
{enum_values_str}
        """.strip()
        )
    all_strs = ["from enum import Enum", *enum_strs]
    enum_str = "\n".join(all_strs)
    return format_str(enum_str, mode=FileMode())


class Column(BaseModel):
    name: str
    data_type: str
    is_nullable: bool
    constraints_info: str | None

    basemodel_type: T.Type[BaseModel] | None = None
    cardinality: Cardinality = Cardinality.ONE

    default_value: str | None = None

    original_column: T.Optional["Column"] = None
    computed_column: T.Optional["ComputedColumn"] = None


class Table(BaseModel):
    name: str
    comment: str | None
    columns: list[Column] = Field(default_factory=list)
    constraints: str | None = None

    def __hash__(self):
        return hash(self.name)


def build_nodes_and_resolvers_str(
    *,
    db_dsn: str,
    base_node: T.Type[Node] | None = None,
    base_resolver: T.Type[Resolver] | None = None,
    configs_by_table_name: dict[str, TableConfig],
    custom_columns_query: str = DEFAULT_COLUMNS_QUERY,
    custom_tables_query: str = DEFAULT_TABLES_QUERY,
) -> str:
    # first get columns
    ordered_columns_by_table_name: dict[str, list[str]] = {}
    tables_by_name: dict[str, Table] = {}
    with psycopg.connect(db_dsn) as conn:
        conn.row_factory = psycopg.rows.dict_row
        with conn.cursor() as cur:
            cur.execute(custom_tables_query)
            for record in cur.fetchall():
                table_name = record["table_name"]
                table = Table(
                    name=table_name,
                    comment=record["comment"],
                    constraints=record["constraints"],
                )
                tables_by_name[table_name] = table
            cur.execute(custom_columns_query)
            for record in cur.fetchall():
                table_name = record["table_name"]
                tables_by_name[table_name].columns.append(
                    Column(
                        name=record["column_name"],
                        data_type=record["data_type"],
                        is_nullable=record["is_nullable"],
                        constraints_info=record["constraints_info"],
                        default_value=record["default_value"],
                    )
                )
                if table_name not in ordered_columns_by_table_name:
                    ordered_columns_by_table_name[table_name] = []
                ordered_columns_by_table_name[table_name].append(record["column_name"])

    mixin_strs: list[str] = []
    node_strs: list[str] = []
    insert_strs: list[str] = []
    patch_strs: list[str] = []
    resolver_strs: list[str] = []

    if base_node:
        base_node_name = f"{base_node.__name__}__"
        mixin_strs.append(
            f"from {base_node.__module__} import {base_node.__name__} as {base_node_name}"
        )
    else:
        base_node_name = "Node"

    if base_resolver:
        base_resolver_name = f"{base_resolver.__name__}__"
        mixin_strs.append(
            f"from {base_resolver.__module__} import {base_resolver.__name__} as {base_resolver_name}"
        )
    else:
        base_resolver_name = "Resolver"

    for table in tables_by_name.values():
        # table config first
        table_config = configs_by_table_name.get(table.name)
        base_node_import_name: str | None = None
        base_resolver_import_name: str | None = None
        computed_columns_str = (
            "\t__computed_columns__: T.ClassVar[dict[str, ComputedColumn]] = {}"
        )
        table_name_str = f'\t__table_name__: T.ClassVar[str] = "{table.name}"'
        ordered_cols_str = ", ".join(
            [f'"{f}"' for f in ordered_columns_by_table_name[table.name]]
        )
        columns_in_order_str = (
            f"\t__columns_in_order__: T.ClassVar[list[str]] = [{ordered_cols_str}]"
        )
        computed_links: dict[str, ComputedLink] = {}
        if table_config:
            if table_config.base_node:
                if not hasattr(table_config.base_node, "__computed_columns__"):
                    table_config.base_node.__computed_columns__ = {}
                if not hasattr(table_config.base_node, "__computed_links__"):
                    table_config.base_node.__computed_links__ = {}
                computed_links = table_config.base_node.__computed_links__
                computed_columns_str = ""
                base_node_import_name = f"{table_config.base_node.__name__}__"
                mixin_strs.append(
                    f"from {table_config.base_node.__module__} import {table_config.base_node.__name__} as {base_node_import_name}"
                )
                # get the computed and default columns
                columns_by_name: dict[str, Column] = {c.name: c for c in table.columns}
                for (
                    computed_name,
                    computed_c,
                ) in table_config.base_node.__computed_columns__.items():
                    original_column: Column | None = None
                    computed_c: ComputedColumn
                    # computed names will overwrite given ones
                    if computed_name in columns_by_name:
                        if not computed_c.override:
                            raise GeneratorError(
                                f"Computed colum {computed_name} is already a column. "
                                f"If you want to override it, set override=True on the ComputedColumn."
                            )
                        col_to_remove = columns_by_name[computed_name]
                        original_column = col_to_remove
                        table.columns.remove(col_to_remove)

                    if computed_c.basemodel_type:
                        import_s = f"from {computed_c.basemodel_type.__module__} import {computed_c.basemodel_type.__name__} as {computed_c.basemodel_type.__name__}__"
                        if import_s not in mixin_strs:
                            mixin_strs.append(import_s)

                    table.columns.append(
                        Column(
                            name=computed_name,
                            data_type=computed_c.data_type,
                            is_nullable=computed_c.is_nullable,
                            constraints_info=None,
                            basemodel_type=computed_c.basemodel_type,
                            cardinality=computed_c.cardinality,
                            original_column=original_column,
                            computed_column=computed_c,
                        )
                    )
            if table_config.base_resolver:
                base_resolver_import_name = f"{table_config.base_resolver.__name__}__"
                mixin_strs.append(
                    f"from {table_config.base_resolver.__module__} import {table_config.base_resolver.__name__} as {base_resolver_import_name}"
                )

        safe_table_name = convert_to_variable_name(table.name)
        get_params: set[str] = set()
        primary_keys: list[str] = []

        # first, add composite keys to get params
        if table.constraints:
            matches = re.findall(
                r"PRIMARY KEY \(PRIMARY KEY \((.*?)\)", table.constraints
            )
            if matches:
                if len(matches) > 1:
                    raise GeneratorError(
                        f"Found more than one primary key for table {table.name}."
                    )
                primary_keys = matches[0].split(", ")

        inner_primary_keys = [f'"{p}"' for p in primary_keys]
        primary_keys_str = f"\t__primary_keys__: T.ClassVar[list[str]] = [{', '.join(inner_primary_keys)}]"

        col_names_to_refs: dict[str, Ref] = {}
        col_strs: list[str] = []
        insert_fields_strs: list[str] = []
        patch_fields_strs: list[str] = []
        filter_by_strs: list[str] = []
        filter_in_strs: list[str] = []
        col_name_to_anno: dict[str, str] = {}
        for col in table.columns:
            if col.data_type.endswith("[]"):
                is_list = True
                data_type = col.data_type.replace("[]", "")
            else:
                is_list = False
                data_type = col.data_type
            if col.basemodel_type:
                python_anno = f"{col.basemodel_type.__name__}__"
            else:
                if data_type in TYPES_MAP:
                    python_anno = TYPES_MAP[data_type].name
                else:
                    # this is an enum
                    python_anno = f"enums.{data_type}"
            if is_list:
                python_anno = f"list[{python_anno}]"
            if col.cardinality == Cardinality.MANY:
                python_anno = f"list[{python_anno}]"
            if col.is_nullable:
                python_anno = f"{python_anno} | None"
            col_strs.append(f"{col.name}: {python_anno} = None")
            if not col.basemodel_type:
                filter_by_strs.append(
                    f"{col.name}: T.Union[{python_anno}, UNSET] = UNSET"
                )
                filter_in_strs.append(
                    f"{col.name}: T.Union[list[{python_anno}], UNSET] = UNSET"
                )
            col_name_to_anno[col.name] = python_anno
            # handle constraints now
            if col.constraints_info:
                constraints = [c.strip() for c in col.constraints_info.split(";")]
                for constraint in constraints:
                    if constraint.startswith("UNIQUE:") or constraint.startswith(
                        "UNIQUE INDEX:"
                    ):
                        if not col.is_nullable:
                            get_params.add(f"{col.name}: {python_anno} = None")
                    elif constraint.startswith("FOREIGN KEY:"):
                        col_names_to_refs[col.name] = Ref.from_str(
                            s=constraint[constraint.index("references ") + 11 :],
                            is_nullable=col.is_nullable,
                        )
                    elif constraint.startswith("PRIMARY KEY:"):
                        # handling this later
                        continue
                    else:
                        raise GeneratorError(f"Invalid constraint: {constraint=}")

            python_anno_insert = f"T.Union[{python_anno}, raw_sql]"
            python_anno_patch = f"T.Optional[T.Union[{python_anno}, raw_sql]]"

            # if computed, ignore
            # if computed and has original col, keep
            if (
                table_config
                and table_config.base_node
                and col.name in table_config.base_node.__computed_columns__
            ):
                if col.original_column:
                    insert_fields_strs.append(
                        f'{col.name}: {python_anno_insert}{" = None" if col.original_column.is_nullable or col.original_column.default_value else ""}'
                    )
            else:
                insert_fields_strs.append(
                    f'{col.name}: {python_anno_insert}{" = None" if col.is_nullable or col.default_value else ""}'
                )
            if col.name not in primary_keys:
                patch_fields_strs.append(f"{col.name}: {python_anno_patch} = None")

        # now add get params for primary keys
        if primary_keys:
            if len(primary_keys) == 1:
                primary_key = primary_keys[0]
                get_params.add(f"{primary_key}: {col_name_to_anno[primary_key]} = None")
            else:
                combined_anno_strs: list[str] = []
                for primary_key in primary_keys:
                    combined_anno_strs.append(col_name_to_anno[primary_key])
                get_params.add(
                    f'{"__".join(primary_keys)}: tuple[{", ".join(combined_anno_strs)}] = None'
                )

        # build filter params
        filter_by_strs.sort()
        filter_in_strs.sort()

        filter_by_params_str = ", ".join(filter_by_strs)
        filter_by_inner_params = ", ".join(
            [f"{n.split(':')[0]}={n.split(':')[0]}" for n in filter_by_strs]
        )

        filter_in_params_str = ", ".join(filter_in_strs)
        filter_in_inner_params = ", ".join(
            [f"{n.split(':')[0]}={n.split(':')[0]}" for n in filter_by_strs]
        )

        # build resolver functions like artist()
        # AND build node functions like artist() -> Artist
        resolver_func_strs: list[str] = []
        node_func_strs: list[str] = []
        for c_name, ref in col_names_to_refs.items():
            safe_c_name = convert_to_variable_name(c_name)
            c_name_key = f"_{ref.column_name}"
            if c_name_key in safe_c_name:
                safe_c_name = safe_c_name[: safe_c_name.rindex(f"_{ref.column_name}")]
            else:
                safe_c_name = f"{safe_c_name}_"
            safe_target_table_name = convert_to_variable_name(ref.table_name)
            safe_rez_name = f"{safe_target_table_name}Resolver"
            resolver_func_strs.append(
                f"""
    def {safe_c_name}(self, resolver: T.Optional["{safe_rez_name}"] = None, key: str = "{safe_c_name}") -> "{safe_table_name}Resolver":
        self.add_child(
            key=key,
            resolver = resolver or {safe_rez_name}(),
            cardinality=Cardinality.ONE,
            from_='FROM "{ref.table_name}" $current WHERE $current.{ref.column_name} = $parent.{c_name}',
            is_required={'True' if not ref.is_nullable else 'False'}
        )
        return self
            """
            )
            node_func_strs.append(
                f'def {safe_c_name}(self, key: str = "{safe_c_name}") -> "{safe_target_table_name}": return self._resolve_node(key=key)'
            )
        for link_name, link in computed_links.items():
            target_rez_name = f"{link.node_name}Resolver"
            kwargs_strs: list[str] = []
            if link.annotations_by_kwarg:
                for var_name, var_anno in link.annotations_by_kwarg.items():
                    kwargs_strs.append(f"{var_name}: {var_anno}")

            kwargs_str = ""
            update_qb_str = ""
            if kwargs_strs:
                kwargs_str = ", *, " + ", ".join(kwargs_strs)
            if link.update_qb:
                if link.annotations_by_kwarg:
                    names_str = ", " + ", ".join(
                        [f"{n}={n}" for n in link.annotations_by_kwarg.keys()]
                    )
                else:
                    names_str = ""
                update_qb_str = f"self._node.__computed_links__['{link_name}'].update_qb(self._qb{names_str})"
            resolver_func_strs.append(
                f"""
    def {link_name}(self, resolver: T.Optional["{target_rez_name}"] = None, key: str = "{link_name}"{kwargs_str}) -> "{safe_table_name}Resolver":
        self.add_child(
            key=key,
            resolver = resolver or {target_rez_name}(),
            cardinality={'Cardinality.ONE' if link.cardinality == Cardinality.ONE else 'Cardinality.MANY'},
            from_='{link.from_}',
            is_required={'False' if link.is_nullable else 'True'}
        )
        {update_qb_str}
        return self
            """
            )
            target_node_name = (
                f'"{link.node_name}"'
                if link.cardinality == Cardinality.ONE
                else f'list["{link.node_name}"]'
            )
            if link.is_nullable:
                target_node_name = f"T.Optional[{target_node_name}]"
            node_func_strs.append(
                f'def {link_name}(self, key: str = "{link_name}") -> {target_node_name}: return self._resolve_node(key=key)'
            )

        resolver_func_str = "\n".join(sorted(resolver_func_strs))
        node_func_str = "\n".join(sorted([f"\t{n}" for n in node_func_strs]))

        get_col_names = [p.split(":")[0] for p in get_params]
        get_inner_params = ", ".join(sorted([f"{n}={n}" for n in get_col_names]))
        col_str = "\n".join(sorted([f"\t{c}" for c in col_strs]))

        node_strs.append(
            f"""
class {safe_table_name}({base_node_import_name or base_node_name}):
{col_str}
{node_func_str}
{computed_columns_str}
{columns_in_order_str}
{table_name_str}
{primary_keys_str}
        """
        )
        insert_str = "\n".join(sorted([f"\t{c}" for c in insert_fields_strs]))
        insert_strs.append(
            f"""
class {safe_table_name}Insert(Insert):
{insert_str}
            """
        )
        if patch_fields_strs:
            patch_str = "\n".join(sorted([f"\t{c}" for c in patch_fields_strs]))
            patch_strs.append(
                f"""
class {safe_table_name}Patch(Patch):
{patch_str}
                """
            )

        if get_params:
            joined_get_params = ", ".join(sorted(get_params))
            get_strs = f"""
    async def get(self, conn: AsyncConnection | None = None, *, {joined_get_params}) -> {safe_table_name} | None:
        return await self._get(conn=conn, {get_inner_params})
    
    async def gerror(self, conn: AsyncConnection | None = None, *, {joined_get_params}) -> {safe_table_name}:
        return await self._gerror(conn=conn, {get_inner_params})
    
    async def query(self, conn: AsyncConnection | None = None, *, format_sql: bool = False, log_query: bool = False) -> list[{safe_table_name}]:
        return await self._query(conn=conn, format_sql=format_sql, log_query=log_query)
    
    async def first_or_none(self, conn: AsyncConnection | None = None, *, format_sql: bool = False, log_query: bool = False) -> {safe_table_name} | None:
        return await super().first_or_none(conn=conn, format_sql=format_sql, log_query=log_query)
    
    async def first(self, conn: AsyncConnection | None = None, *, format_sql: bool = False, log_query: bool = False) -> {safe_table_name}:
        return await super().first(conn=conn, format_sql=format_sql, log_query=log_query)
    
    async def one_or_none(self, conn: AsyncConnection | None = None, *, format_sql: bool = False, log_query: bool = False) -> {safe_table_name} | None:
        return await super().one_or_none(conn=conn, format_sql=format_sql, log_query=log_query)

    async def one(self, conn: AsyncConnection | None = None, *, format_sql: bool = False, log_query: bool = False) -> {safe_table_name}:
        return await super().one(conn=conn, format_sql=format_sql, log_query=log_query)
    
    def filter_by(self, *, {filter_by_params_str}) -> "{safe_table_name}Resolver":
        return self._filter_by({filter_by_inner_params})
    
    def filter_in(self, *, {filter_in_params_str}) -> "{safe_table_name}Resolver":
        return self._filter_in({filter_in_inner_params})
    
    async def insert_one(self, insert: {safe_table_name}Insert, *, format_sql: bool = False, log_query: bool = False, conn: AsyncConnection | None = None, commit_after_insert: bool = False, force_refresh: bool = False, conflict: Conflict | None = None) -> {safe_table_name}:
        return await super().insert_one(insert=insert, format_sql=format_sql, log_query=log_query, conn=conn, commit_after_insert=commit_after_insert, force_refresh=force_refresh, conflict=conflict)
    
    async def insert_one_or_none(self, insert: {safe_table_name}Insert, *, format_sql: bool = False, log_query: bool = False, conn: AsyncConnection | None = None, commit_after_insert: bool = False, force_refresh: bool = False, conflict: Conflict | None = None) -> T.Optional[{safe_table_name}]:
        return await super().insert_one_or_none(insert=insert, format_sql=format_sql, log_query=log_query, conn=conn, commit_after_insert=commit_after_insert, force_refresh=force_refresh, conflict=conflict)

    async def insert_many(self, inserts: list[{safe_table_name}Insert], conflict: Conflict | None = None, *, format_sql: bool = False, log_query: bool = False, conn: AsyncConnection | None = None, commit_after_insert: bool = False, force_refresh: bool = False) -> list[{safe_table_name}]:
        return await super().insert_many(inserts=inserts, conflict=conflict, format_sql=format_sql, log_query=log_query, conn=conn, commit_after_insert=commit_after_insert, force_refresh=force_refresh)
    
    async def delete_one_or_none(self, *, format_sql: bool = False, log_query: bool = False, conn: AsyncConnection | None = None, commit_after_delete: bool = False) -> T.Optional[{safe_table_name}]:
        return await super().delete_one_or_none(format_sql=format_sql, log_query=log_query, conn=conn, commit_after_delete=commit_after_delete)
    
    async def delete_one(self, *, format_sql: bool = False, log_query: bool = False, conn: AsyncConnection | None = None, commit_after_delete: bool = False) -> {safe_table_name}:
        return await super().delete_one(format_sql=format_sql, log_query=log_query, conn=conn, commit_after_delete=commit_after_delete)
    
    async def delete_many(self, *, format_sql: bool = False, log_query: bool = False, conn: AsyncConnection | None = None, commit_after_delete: bool = False, delete_all: bool = False) -> list[{safe_table_name}]:
        return await super().delete_many(format_sql=format_sql, log_query=log_query, conn=conn, commit_after_delete=commit_after_delete, delete_all=delete_all)
            """
            if patch_fields_strs:
                patch_one_str = f"""
    async def patch_one(self, patch: {safe_table_name}Patch, *, format_sql: bool = False, log_query: bool = False, conn: AsyncConnection | None = None, commit_after_patch: bool = False, force_refresh: bool = False) -> {safe_table_name}:
        return await super().patch_one(patch=patch, format_sql=format_sql, log_query=log_query, conn=conn, commit_after_patch=commit_after_patch, force_refresh=force_refresh)
    
    async def patch_many(self, patch: {safe_table_name}Patch, *, format_sql: bool = False, log_query: bool = False, conn: AsyncConnection | None = None, commit_after_patch: bool = False, force_refresh: bool = False, patch_all: bool = False) -> list[{safe_table_name}]:
        return await super().patch_many(patch=patch, format_sql=format_sql, log_query=log_query, conn=conn, commit_after_patch=commit_after_patch, force_refresh=force_refresh, patch_all=patch_all)
                """
                get_strs = f"{get_strs}\n{patch_one_str}"
        else:
            get_strs = ""

        include_col_anno_strs: list[str] = []
        include_col_inner_strs: list[str] = []
        if table.columns:
            for col in table.columns:
                computed_c = col.computed_column
                if computed_c and computed_c.include_args_annotation:
                    anno = computed_c.include_args_annotation
                    if computed_c.is_nullable:
                        anno = f"T.Optional[{anno}]"
                    include_col_anno_strs.append(f"{col.name}: {anno} = UNSET")
                else:
                    include_col_anno_strs.append(f"{col.name}: bool = False")
                include_col_inner_strs.append(f"{col.name}={col.name}")

        if include_col_anno_strs:
            include_col_str: str = ", ".join(sorted(include_col_anno_strs))
            include_col_inner_strs: str = ", ".join(sorted(include_col_inner_strs))
            include_func_str = f'def include(self, *, {include_col_str}) -> "{safe_table_name}Resolver": return self._include({include_col_inner_strs})'
        else:
            include_func_str = ""

        resolver_strs.append(
            f"""
class {safe_table_name}Resolver({base_resolver_import_name or base_resolver_name}):
    _node = {safe_table_name}
    
{get_strs}
{resolver_func_str}
    {include_func_str}
        """
        )
    all_strs = [
        "import typing as T",
        "import datetime",
        "from uuid import UUID",
        "from pydantic_core import PydanticUndefined as UNSET",
        "from sqlalchemy.ext.asyncio import AsyncConnection",
        "from asqlorm.models import Node, Insert, Patch, Resolver, Conflict, ComputedColumn, raw_sql",
        "from fastgql.query_builders.sql.query_builder import Cardinality",
        "from . import enums",
        *mixin_strs,
        *sorted(node_strs),
        *sorted(insert_strs),
        *sorted(patch_strs),
        *sorted(resolver_strs),
    ]
    all_str = "\n".join(all_strs)
    return format_str(all_str, mode=FileMode())



import typing as T
from uuid import UUID
from datetime import date, datetime, time
from pydantic import BaseModel


class TypeData(BaseModel):
    name: str
    val: T.Type
    serializer: T.Callable


TYPES_MAP: dict[str, TypeData] = {
    "uuid": TypeData(name="UUID", val=UUID, serializer=UUID),
    "double precision": TypeData(name="float", val=float, serializer=float),
    "float": TypeData(name="float", val=float, serializer=float),
    "date": TypeData(name="datetime.date", val=date, serializer=date.fromisoformat),
    "timestamp with time zone": TypeData(
        name="datetime.datetime", val=datetime, serializer=datetime.fromisoformat
    ),
    "timestamp": TypeData(
        name="datetime.datetime", val=datetime, serializer=datetime.fromisoformat
    ),
    "datetime": TypeData(
        name="datetime.datetime", val=datetime, serializer=datetime.fromisoformat
    ),
    "time": TypeData(name="datetime.time", val=time, serializer=time.fromisoformat),
    "time without time zone": TypeData(
        name="datetime.time", val=time, serializer=time.fromisoformat
    ),
    "character varying": TypeData(name="str", val=str, serializer=str),
    "boolean": TypeData(name="bool", val=bool, serializer=bool),
    "bool": TypeData(name="bool", val=bool, serializer=bool),
    "integer": TypeData(name="int", val=int, serializer=int),
    "int": TypeData(name="int", val=int, serializer=int),
    "text": TypeData(name="str", val=str, serializer=str),
    "varchar": TypeData(name="str", val=str, serializer=str),
    "str": TypeData(name="str", val=str, serializer=str),
    "jsonb": TypeData(name="dict", val=dict, serializer=dict),
    "json": TypeData(name="dict", val=dict, serializer=dict),
}

__all__ = ["TypeData", "TYPES_MAP"]
